@@ -16,23 +16,21 @@
 #ifndef P_POLYS_H
 #define P_POLYS_H
 
-#include <omalloc/omalloc.h>
+#include "misc/mylimits.h"
+#include "misc/intvec.h"
+#include "coeffs/coeffs.h"
 
-#include <misc/mylimits.h>
-#include <misc/intvec.h>
-#include <coeffs/coeffs.h>
+#include "polys/monomials/monomials.h"
+#include "polys/monomials/ring.h"
 
-#include <polys/monomials/monomials.h>
-#include <polys/monomials/ring.h>
+#include "polys/templates/p_MemAdd.h"
+#include "polys/templates/p_MemCmp.h"
+#include "polys/templates/p_Procs.h"
 
-#include <polys/templates/p_MemAdd.h>
-#include <polys/templates/p_MemCmp.h>
-#include <polys/templates/p_Procs.h>
-
-#include <polys/sbuckets.h>
+#include "polys/sbuckets.h"
 
 #ifdef HAVE_PLURAL
-#include <polys/nc/nc.h>
+#include "polys/nc/nc.h"
 #endif
 
 poly p_Farey(poly p, number N, const ring r);
@@ -96,6 +94,10 @@ poly      p_ISet(long i, const ring r);
 poly      p_NSet(number n, const ring r);
 
 void  p_Vec2Polys(poly v, poly**p, int *len, const ring r);
+poly  p_Vec2Poly(poly v, int k, const ring r);
+
+/// julia: vector to already allocated array (len=p_MaxComp(v,r))
+void  p_Vec2Array(poly v, poly *p, int len, const ring r);
 
 /***************************************************************
  *
@@ -207,10 +209,12 @@ void      p_Norm(poly p1, const ring r);
 void      p_Normalize(poly p,const ring r);
 void      p_ProjectiveUnique(poly p,const ring r);
 
+void      p_ContentForGB(poly p, const ring r);
 void      p_Content(poly p, const ring r);
 #if 1
 // currently only used by Singular/janet
 void      p_SimpleContent(poly p, int s, const ring r);
+number    p_InitContent(poly ph, const ring r);
 #endif
 
 poly      p_Cleardenom(poly p, const ring r);
@@ -487,10 +491,10 @@ static inline unsigned long p_SetExp(poly p, const unsigned long e, const unsign
   pAssume2((VarOffset >> (24 + 6)) == 0);
 
   // shift e to the left:
-  register int shift = VarOffset >> 24;
+  REGISTER int shift = VarOffset >> 24;
   unsigned long ee = e << shift /*(VarOffset >> 24)*/;
   // find the bits in the exponent vector
-  register int offset = (VarOffset & 0xffffff);
+  REGISTER int offset = (VarOffset & 0xffffff);
   // clear the bits in the exponent vector:
   p->exp[offset]  &= ~( iBitmask << shift );
   // insert e with |
@@ -531,12 +535,12 @@ static inline long p_SetExp(poly p, const long e, const unsigned long iBitmask,
   pAssume2(e <= BitMask(iBitmask, VarOffset >> 30));
 
   // shift e to the left:
-  register int hbyte = VarOffset >> 24;
+  REGISTER int hbyte = VarOffset >> 24;
   int bitmask = BitMask(iBitmask, hbyte >> 6);
-  register int shift = hbyte & 0x3f;
+  REGISTER int shift = hbyte & 0x3f;
   long ee = e << shift;
   // find the bits in the exponent vector
-  register int offset = (VarOffset & 0xffffff);
+  REGISTER int offset = (VarOffset & 0xffffff);
   // clear the bits in the exponent vector:
   p->exp[offset]  &= ~( bitmask << shift );
   // insert e with |
@@ -796,19 +800,28 @@ p_GetTotalDegree(const unsigned long l, const ring r, const int number_of_exps)
 /// returns a copy of p (without any additional testing)
 static inline poly p_Copy_noCheck(poly p, const ring r)
 {
-  assume(r != NULL); assume(r->p_Procs != NULL); assume(r->p_Procs->p_Copy != NULL);
+  /*assume(p!=NULL);*/
+  assume(r != NULL);
+  assume(r->p_Procs != NULL);
+  assume(r->p_Procs->p_Copy != NULL);
   return r->p_Procs->p_Copy(p, r);
 }
 
 /// returns a copy of p
 static inline poly p_Copy(poly p, const ring r)
 {
-  p_Test(p,r);
-  const poly pp = p_Copy_noCheck(p, r);
-  p_Test(pp,r);
-  return pp;
+  if (p!=NULL)
+  {
+    p_Test(p,r);
+    const poly pp = p_Copy_noCheck(p, r);
+    p_Test(pp,r);
+    return pp;
+  }
+  else
+    return NULL;
 }
 
+/// copy the i(leading) term of p
 static inline poly p_Head(poly p, const ring r)
 {
   if (p == NULL) return NULL;
@@ -822,7 +835,10 @@ static inline poly p_Head(poly p, const ring r)
   return np;
 }
 
-// returns a copy of p with Lm(p) from lmRing and Tail(p) from tailRing
+/// like p_Head, but with coefficient 1
+poly p_CopyPowerProduct(poly p, const ring r);
+
+/// returns a copy of p with Lm(p) from lmRing and Tail(p) from tailRing
 static inline poly p_Copy(poly p, const ring lmRing, const ring tailRing)
 {
   if (p != NULL)
@@ -832,7 +848,8 @@ static inline poly p_Copy(poly p, const ring lmRing, const ring tailRing)
       return p_Copy_noCheck(p, tailRing);
 #endif
     poly pres = p_Head(p, lmRing);
-    pNext(pres) = p_Copy_noCheck(pNext(p), tailRing);
+    if (pNext(p)!=NULL)
+      pNext(pres) = p_Copy_noCheck(pNext(p), tailRing);
     return pres;
   }
   else
@@ -843,7 +860,8 @@ static inline poly p_Copy(poly p, const ring lmRing, const ring tailRing)
 static inline void p_Delete(poly *p, const ring r)
 {
   assume( p!= NULL );
-  r->p_Procs->p_Delete(p, r);
+  assume( r!= NULL );
+  if ((*p)!=NULL) r->p_Procs->p_Delete(p, r);
 }
 
 static inline void p_Delete(poly *p,  const ring lmRing, const ring tailRing)
@@ -865,7 +883,7 @@ static inline void p_Delete(poly *p,  const ring lmRing, const ring tailRing)
 }
 
 // copys monomials of p, allocates new monomials from bin,
-// deletes monomoals of p
+// deletes monomials of p
 static inline poly p_ShallowCopyDelete(poly p, const ring r, omBin bin)
 {
   p_LmCheckPolyRing2(p, r);
@@ -877,6 +895,8 @@ static inline poly p_ShallowCopyDelete(poly p, const ring r, omBin bin)
 static inline poly p_Add_q(poly p, poly q, const ring r)
 {
   assume( (p != q) || (p == NULL && q == NULL) );
+  if (q==NULL) return p;
+  if (p==NULL) return q;
   int shorter;
   return r->p_Procs->p_Add_q(p, q, shorter, r);
 }
@@ -885,28 +905,34 @@ static inline poly p_Add_q(poly p, poly q, const ring r)
 static inline poly p_Add_q(poly p, poly q, int &lp, int lq, const ring r)
 {
   assume( (p != q) || (p == NULL && q == NULL) );
+  if (q==NULL) return p;
+  if (p==NULL) { lp=lq; return q; }
   int shorter;
   poly res = r->p_Procs->p_Add_q(p, q, shorter, r);
-  lp = (lp + lq) - shorter;
+  lp += lq - shorter;
   return res;
 }
 
 // returns p*n, destroys p
 static inline poly p_Mult_nn(poly p, number n, const ring r)
 {
+  if (p==NULL) return NULL;
   if (n_IsOne(n, r->cf))
     return p;
   else if (n_IsZero(n, r->cf))
   {
-    r->p_Procs->p_Delete(&p, r); // NOTE: without p_Delete - memory leak!
+    p_Delete(&p, r); // NOTE: without p_Delete - memory leak!
     return NULL;
-  } else
+  }
+  else
     return r->p_Procs->p_Mult_nn(p, n, r);
 }
+#define __p_Mult_nn(p,n,r) r->p_Procs->p_Mult_nn(p, n, r)
 
 static inline poly p_Mult_nn(poly p, number n, const ring lmRing,
                         const ring tailRing)
 {
+  assume(p!=NULL);
 #ifndef PDEBUG
   if (lmRing == tailRing)
     return p_Mult_nn(p, n, tailRing);
@@ -914,18 +940,25 @@ static inline poly p_Mult_nn(poly p, number n, const ring lmRing,
   poly pnext = pNext(p);
   pNext(p) = NULL;
   p = lmRing->p_Procs->p_Mult_nn(p, n, lmRing);
-  pNext(p) = tailRing->p_Procs->p_Mult_nn(pnext, n, tailRing);
+  if (pnext!=NULL)
+  {
+    pNext(p) = tailRing->p_Procs->p_Mult_nn(pnext, n, tailRing);
+  }
   return p;
 }
 
 // returns p*n, does not destroy p
 static inline poly pp_Mult_nn(poly p, number n, const ring r)
 {
+  if (p==NULL) return NULL;
   if (n_IsOne(n, r->cf))
     return p_Copy(p, r);
+  else if (n_IsZero(n, r->cf))
+    return NULL;
   else
     return r->p_Procs->pp_Mult_nn(p, n, r);
 }
+#define __pp_Mult_nn(p,n,r) r->p_Procs->pp_Mult_nn(p, n, r)
 
 // test if the monomial is a constant as a vector component
 // i.e., test if all exponents are zero
@@ -956,19 +989,19 @@ static inline BOOLEAN p_LmIsConstant(const poly p, const ring r)
 // returns Copy(p)*m, does neither destroy p nor m
 static inline poly pp_Mult_mm(poly p, poly m, const ring r)
 {
+  if (p==NULL) return NULL;
   if (p_LmIsConstant(m, r))
-    return pp_Mult_nn(p, pGetCoeff(m), r);
+    return __pp_Mult_nn(p, pGetCoeff(m), r);
   else
-  {
     return r->p_Procs->pp_Mult_mm(p, m, r);
-  }
 }
 
 // returns p*m, destroys p, const: m
 static inline poly p_Mult_mm(poly p, poly m, const ring r)
 {
+  if (p==NULL) return NULL;
   if (p_LmIsConstant(m, r))
-    return p_Mult_nn(p, pGetCoeff(m), r);
+    return __p_Mult_nn(p, pGetCoeff(m), r);
   else
     return r->p_Procs->p_Mult_mm(p, m, r);
 }
@@ -1023,43 +1056,30 @@ static inline poly p_Mult_q(poly p, poly q, const ring r)
 
   if (p == NULL)
   {
-    r->p_Procs->p_Delete(&q, r);
+    p_Delete(&q, r);
     return NULL;
   }
   if (q == NULL)
   {
-    r->p_Procs->p_Delete(&p, r);
+    p_Delete(&p, r);
     return NULL;
   }
 
   if (pNext(p) == NULL)
   {
-#ifdef HAVE_PLURAL
-    if (rIsPluralRing(r))
-      q = nc_mm_Mult_p(p, q, r);
-    else
-#endif /* HAVE_PLURAL */
-      q = r->p_Procs->p_Mult_mm(q, p, r);
-
-    r->p_Procs->p_Delete(&p, r);
+    q = r->p_Procs->p_mm_Mult(q, p, r);
+    p_LmDelete(&p, r);
     return q;
   }
 
   if (pNext(q) == NULL)
   {
-  // NEEDED
-#ifdef HAVE_PLURAL
-/*    if (rIsPluralRing(r))
-      p = gnc_p_Mult_mm(p, q, r); // ???
-    else*/
-#endif /* HAVE_PLURAL */
-      p = r->p_Procs->p_Mult_mm(p, q, r);
-
-    r->p_Procs->p_Delete(&q, r);
+    p = r->p_Procs->p_Mult_mm(p, q, r);
+    p_LmDelete(&q, r);
     return p;
   }
 #ifdef HAVE_PLURAL
-  if (rIsPluralRing(r))
+  if (rIsNCRing(r))
     return _nc_p_Mult_q(p, q, r);
   else
 #endif
@@ -1073,11 +1093,7 @@ static inline poly pp_Mult_qq(poly p, poly q, const ring r)
 
   if (pNext(p) == NULL)
   {
-#ifdef HAVE_PLURAL
-    if (rIsPluralRing(r))
-      return nc_mm_Mult_pp(p, q, r);
-#endif
-    return r->p_Procs->pp_Mult_mm(q, p, r);
+    return r->p_Procs->pp_mm_Mult(q, p, r);
   }
 
   if (pNext(q) == NULL)
@@ -1448,13 +1464,30 @@ static inline void p_GetExpV(poly p, int *ev, const ring r)
 
   ev[0] = p_GetComp(p, r);
 }
+// p_GetExpVL is used in Singular,jl
+static inline void p_GetExpVL(poly p, int64 *ev, const ring r)
+{
+  p_LmCheckPolyRing1(p, r);
+  for (unsigned j = r->N; j!=0; j--)
+      ev[j-1] = p_GetExp(p, j, r);
+}
 static inline void p_SetExpV(poly p, int *ev, const ring r)
 {
   p_LmCheckPolyRing1(p, r);
   for (unsigned j = r->N; j!=0; j--)
       p_SetExp(p, j, ev[j], r);
 
-  p_SetComp(p, ev[0],r);
+  if(ev[0]!=0) p_SetComp(p, ev[0],r);
+  p_Setm(p, r);
+}
+// p_SetExpVL is used in Singular,jl
+static inline void p_SetExpVL(poly p, int64 *ev, const ring r)
+{
+  p_LmCheckPolyRing1(p, r);
+  for (unsigned j = r->N; j!=0; j--)
+      p_SetExp(p, j, ev[j], r);
+
+  if(ev[0]!=0) p_SetComp(p, ev[0],r);
   p_Setm(p, r);
 }
 
@@ -1471,11 +1504,11 @@ static inline int p_LmCmp(poly p, poly q, const ring r)
 
   const unsigned long* _s1 = ((unsigned long*) p->exp);
   const unsigned long* _s2 = ((unsigned long*) q->exp);
-  register unsigned long _v1;
-  register unsigned long _v2;
+  REGISTER unsigned long _v1;
+  REGISTER unsigned long _v2;
   const unsigned long _l = r->CmpL_Size;
 
-  register unsigned long _i=0;
+  REGISTER unsigned long _i=0;
 
   LengthGeneral_OrdGeneral_LoopTop:
   _v1 = _s1[_i];
@@ -1487,14 +1520,21 @@ static inline int p_LmCmp(poly p, poly q, const ring r)
     goto LengthGeneral_OrdGeneral_LoopTop;
   }
   const long* _ordsgn = (long*) r->ordsgn;
+#if 1 /* two variants*/
   if (_v1 > _v2)
   {
-    if (_ordsgn[_i] == 1) return 1;
-    return -1;
+    return _ordsgn[_i];
   }
-  if (_ordsgn[_i] == 1) return -1;
-  return 1;
-
+  return -(_ordsgn[_i]);
+#else
+   if (_v1 > _v2)
+   {
+     if (_ordsgn[_i] == 1) return 1;
+     return -1;
+   }
+   if (_ordsgn[_i] == 1) return -1;
+   return 1;
+#endif
 }
 
 // The coefficient will be compared in absolute value
@@ -1615,12 +1655,27 @@ BOOLEAN p_ComparePolys(poly p1,poly p2, const ring r);
 static inline int p_Cmp(poly p1, poly p2, ring r)
 {
   if (p2==NULL)
+  {
+    if (p1==NULL) return 0;
     return 1;
+  }
   if (p1==NULL)
     return -1;
   return p_LmCmp(p1,p2,r);
 }
 
+static inline int p_CmpPolys(poly p1, poly p2, ring r)
+{
+  if (p2==NULL)
+  {
+    if (p1==NULL) return 0;
+    return 1;
+  }
+  if (p1==NULL)
+    return -1;
+  return p_ComparePolys(p1,p2,r);
+}
+
 
 /***************************************************************
  *
@@ -1866,7 +1921,7 @@ static inline BOOLEAN p_LmShortDivisibleBy(poly a, unsigned long sev_a, const ri
  ***************************************************************/
 
 
-// like the respective p_LmIs* routines, except that p might be empty
+/// like the respective p_LmIs* routines, except that p might be empty
 static inline BOOLEAN p_IsConstantComp(const poly p, const ring r)
 {
   if (p == NULL) return TRUE;
@@ -1929,9 +1984,10 @@ static inline BOOLEAN p_LmExpVectorAddIsOk(const poly p1, const poly p2,
 }
 void      p_Split(poly p, poly * r);   /*p => IN(p), r => REST(p) */
 BOOLEAN p_HasNotCF(poly p1, poly p2, const ring r);
+BOOLEAN p_HasNotCFRing(poly p1, poly p2, const ring r);
 poly      p_mInit(const char *s, BOOLEAN &ok, const ring r); /* monom s -> poly, interpreter */
 const char *    p_Read(const char *s, poly &p,const ring r); /* monom -> poly */
-poly      p_Divide(poly a, poly b, const ring r);
+poly      p_MDivide(poly a, poly b, const ring r);
 poly      p_DivideM(poly a, poly b, const ring r);
 poly      p_Div_nn(poly p, const number n, const ring r);
 
