@@ -8,27 +8,19 @@
 */
 
 /* includes */
-#include <omalloc/omalloc.h>
-#include <misc/auxiliary.h>
-#include <coeffs/coeffs.h>
-//#include <polys/monomials/polys-impl.h>
+#include "misc/auxiliary.h"
+#include "coeffs/coeffs.h"
+#include "misc/intvec.h"
+#include "misc/int64vec.h"
+#include "coeffs/coeffs.h" // ring,number
+#include "polys/monomials/monomials.h"
+//#include "polys/monomials/polys-impl.h"
 //
 
 /* forward declaration of types */
 class idrec; typedef idrec *   idhdl; // _only_ for idhdl ip_sring::idroot
-struct  spolyrec;
-typedef struct spolyrec    polyrec;
-typedef struct spolyrec *         poly;
-typedef struct spolyrec const *   const_poly;
-struct ip_sring;
-typedef struct ip_sring *         ring;
-typedef struct ip_sring const *   const_ring;
-class intvec;
-class int64vec;
 struct p_Procs_s;
 typedef struct p_Procs_s p_Procs_s;
-//class slists;
-//typedef slists *           lists;
 class kBucket;
 typedef kBucket*           kBucket_pt;
 
@@ -258,7 +250,7 @@ struct ip_sring
 // general ordering: pointer/structs, long, int, short, BOOLEAN/char/enum
 // general defining procedures: rInit, rComplete, interpreter, ??
   idhdl      idroot; /* local objects , interpreter*/
-  int*       order;  /* array of orderings, rInit/rSleftvOrdering2Ordering */
+  rRingOrder_t* order;  /* array of orderings, rInit/rSleftvOrdering2Ordering */
   int*       block0; /* starting pos., rInit/rSleftvOrdering2Ordering*/
   int*       block1; /* ending pos., rInit/rSleftvOrdering2Ordering*/
 //  char**     parameter; /* names of parameters, rInit */
@@ -328,7 +320,7 @@ struct ip_sring
   BOOLEAN   MixedOrder; // TRUE for global/local mixed orderings, FALSE otherwise
   BOOLEAN   pLexOrder; /* TRUE if the monomial ordering is not compatible with pFDeg */
 
-  BOOLEAN   ComponentOrder; // ???
+  BOOLEAN   ComponentOrder; // 1 if ringorder_c, -1 for ringorder_C,ringorder_S,ringorder_s
 
   // what follows below here should be set by rComplete, _only_
   // contains component, but no weight fields in E */
@@ -387,8 +379,9 @@ struct ip_sring
 
 ring   rDefault(int ch, int N, char **n);
 ring   rDefault(const coeffs cf, int N, char **n, const rRingOrder_t o=ringorder_lp);
-ring   rDefault(int ch, int N, char **n,int ord_size, int *ord, int *block0, int *block1, int **wvhdl=NULL);
-ring   rDefault(const coeffs cf, int N, char **n,int ord_size, int *ord, int *block0, int *block1, int **wvhdl=NULL);
+ring   rDefault(int ch, int N, char **n,int ord_size, rRingOrder_t *ord, int *block0, int *block1, int **wvhdl=NULL);
+ring   rDefault(const coeffs cf, int N, char **n,int ord_size, rRingOrder_t *ord, int *block0, int *block1, int **wvhdl=NULL, unsigned long bitmask=0);
+unsigned long rGetExpSize(unsigned long bitmask, int & bits, int N);
 
 // #define rIsRingVar(A) r_IsRingVar(A,currRing)
 int    r_IsRingVar(const char *n, char**names, int N);
@@ -403,10 +396,29 @@ ring   rEnvelope(ring r);
 /// we must always have this test!
 static inline BOOLEAN rIsPluralRing(const ring r)
 {
-  assume(r != NULL); assume(r->cf != NULL);
+  assume(r != NULL);
 #ifdef HAVE_PLURAL
   nc_struct *n;
-  return (r != NULL) && ((n=r->GetNC()) != NULL) /*&& (n->type != nc_error)*/;
+  return ((n=r->GetNC()) != NULL) /*&& (n->type != nc_error)*/;
+#else
+  return FALSE;
+#endif
+}
+
+static inline BOOLEAN rIsLPRing(const ring r)
+{
+  assume(r != NULL);
+#if defined(HAVE_PLURAL) && defined(HAVE_SHIFTBBA)
+  return (r->isLPring!=0);
+#else
+  return FALSE;
+#endif
+}
+
+static inline BOOLEAN rIsNCRing(const ring r)
+{
+#ifdef HAVE_PLURAL
+  return rIsLPRing(r) || rIsPluralRing(r);
 #else
   return FALSE;
 #endif
@@ -431,7 +443,7 @@ void rGetSComps(int** currComponents, long** currShiftedComponents, int *length,
 
 
 const char * rSimpleOrdStr(int ord);
-int rOrderName(char * ordername);
+rRingOrder_t rOrderName(char * ordername);
 char * rOrdStr(ring r);
 char * rVarStr(ring r);
 char * rCharStr(ring r);
@@ -458,22 +470,19 @@ BOOLEAN rSamePolyRep(ring r1, ring r2);
 
 void   rUnComplete(ring r);
 
-BOOLEAN rRing_is_Homog(ring r);
-BOOLEAN rRing_has_CompLastBlock(ring r);
+BOOLEAN rRing_is_Homog(const ring r);
+BOOLEAN rRing_has_CompLastBlock(const ring r);
+BOOLEAN rRing_ord_pure_dp(const ring r);
+BOOLEAN rRing_ord_pure_Dp(const ring r);
+BOOLEAN rRing_ord_pure_lp(const ring r);
 
 #ifdef HAVE_RINGS
 static inline BOOLEAN rField_is_Ring_2toM(const ring r)
 { assume(r != NULL); assume(r->cf != NULL); return ( nCoeff_is_Ring_2toM(r->cf) ); }
 
-static inline BOOLEAN rField_is_Ring_ModN(const ring r)
-{ assume(r != NULL); assume(r->cf != NULL); return ( nCoeff_is_Ring_ModN(r->cf) ); }
-
 static inline BOOLEAN rField_is_Ring_PtoM(const ring r)
 { assume(r != NULL); assume(r->cf != NULL); return ( nCoeff_is_Ring_PtoM(r->cf) ); }
 
-static inline BOOLEAN rField_is_Ring_Z(const ring r)
-{ assume(r != NULL); assume(r->cf != NULL); return ( nCoeff_is_Ring_Z(r->cf) ); }
-
 static inline BOOLEAN rField_is_Ring(const ring r)
 { assume(r != NULL); assume(r->cf != NULL); return nCoeff_is_Ring(r->cf); }
 
@@ -485,9 +494,7 @@ static inline BOOLEAN rField_has_Units(const ring r)
 #else
 #define rField_is_Ring(A) (0)
 #define rField_is_Ring_2toM(A) (0)
-#define rField_is_Ring_ModN(A) (0)
 #define rField_is_Ring_PtoM(A) (0)
-#define rField_is_Ring_Z(A) (0)
 #define rField_is_Domain(A) (1)
 #define rField_has_Units(A) (1)
 #endif
@@ -501,6 +508,12 @@ static inline BOOLEAN rField_is_Zp(const ring r, int p)
 static inline BOOLEAN rField_is_Q(const ring r)
 { assume(r != NULL); assume(r->cf != NULL); return nCoeff_is_Q(r->cf); }
 
+static inline BOOLEAN rField_is_Z(const ring r)
+{ assume(r != NULL); assume(r->cf != NULL); return nCoeff_is_Z(r->cf); }
+
+static inline BOOLEAN rField_is_Zn(const ring r)
+{ assume(r != NULL); assume(r->cf != NULL); return nCoeff_is_Zn(r->cf); }
+
 static inline BOOLEAN rField_is_numeric(const ring r) /* R, long R, long C */
 { assume(r != NULL); assume(r->cf != NULL); return nCoeff_is_numeric(r->cf); }
 
@@ -541,7 +554,8 @@ static inline BOOLEAN rField_has_simple_inverse(const ring r)
 static inline BOOLEAN rField_has_simple_Alloc(const ring r)
 { assume(r != NULL); assume(r->cf != NULL); return nCoeff_has_simple_Alloc(r->cf); }
 
-n_coeffType rFieldType(const ring r);
+/// the type of the coefficient filed of r (n_Zp, n_Q, etc)
+static inline n_coeffType rFieldType(const ring r) { return (r->cf->type); }
 
 /// this needs to be called whenever a new ring is created: new fields
 /// in ring are created (like VarOffset), unless they already exist
@@ -717,6 +731,7 @@ void   rSetSyzComp(int k, const ring r);
 ring   rAssure_HasComp(const ring r);
 ring   rAssure_SyzOrder(const ring r, BOOLEAN complete);
 ring   rAssure_SyzComp(const ring r, BOOLEAN complete = TRUE);
+ring   rAssure_InducedSchreyerOrdering(const ring r, BOOLEAN complete = TRUE, int sgn = 1);
 
 ring   rAssure_dp_S(const ring r);
 ring   rAssure_dp_C(const ring r);
@@ -727,8 +742,8 @@ ring   rAssure_c_dp(const ring r);
 ring   rAssure_CompLastBlock(const ring r, BOOLEAN complete = TRUE);
 
 /// makes sure that c/C ordering is last ordering and SyzIndex is first
-ring   rAssure_SyzComp_CompLastBlock(const ring r, BOOLEAN complete = TRUE);
-ring   rAssure_TDeg(const ring r, int start_var, int end_var, int &pos);
+ring   rAssure_SyzComp_CompLastBlock(const ring r);
+ring   rAssure_TDeg(const ring r, int &pos);
 
 /// return the max-comonent wchich has syzIndex i
 /// Assume: i<= syzIndex_limit
@@ -750,9 +765,6 @@ inline BOOLEAN rHasMixedOrdering(const ring r) { return (r->MixedOrder); }
 // #define rHasGlobalOrdering(R) ((R)->OrdSgn==1)
 // #define rHasLocalOrMixedOrdering(R) ((R)->OrdSgn==-1)
 
-#define rHasGlobalOrdering_currRing() rHasGlobalOrdering(currRing)
-#define rHasLocalOrMixedOrdering_currRing() rHasLocalOrMixedOrdering(currRing)
-
 BOOLEAN rOrd_is_Totaldegree_Ordering(const ring r);
 
 /// return TRUE if p_SetComp requires p_Setm
@@ -791,7 +803,6 @@ void rKillModifiedRing(ring r);
 void rKillModified_Wp_Ring(ring r);
 
 ring rModifyRing_Simple(ring r, BOOLEAN omit_degree, BOOLEAN omit_comp, unsigned long exp_limit, BOOLEAN &simple);
-void rKillModifiedRing_Simple(ring r);
 
 #ifdef RDEBUG
 void rDebugPrint(const ring r);
@@ -814,11 +825,19 @@ BOOLEAN rSetISReference(const ring r, const ideal F, const int i = 0, const int
 
 /// return the position of the p^th IS block order block in r->typ[]...
 int rGetISPos(const int p, const ring r);
+void pISUpdateComponents(ideal F, const intvec *const V, const int MIN, const ring r);
 
 BOOLEAN rCheckIV(const intvec *iv);
 int rTypeOfMatrixOrder(const intvec *order);
 
 void rDelete(ring r); // To be used instead of rKill!
 
-extern omBin sip_sring_bin;
+EXTERN_VAR omBin sip_sring_bin;
+
+// ring manipulation
+/// K[x],"y" -> K[x,y] resp. K[y,x]
+ring rPlusVar(const ring r, char *v,int left);
+
+/// undo rPlusVar
+ring rMinusVar(const ring r, char *v);
 #endif
